from .constants import POSTGRES, SQLITE


async def Connexer(dsn: str, pooled = False):
    vendor, _, remainder = dsn.partition('://')
    if vendor == POSTGRES:
        from asyncpg import connect, create_pool
        if not pooled:
            return vendor, await connect(dsn=dsn)
        return vendor, await create_pool(dsn)
    elif vendor == SQLITE:
        from aiosqlite import connect, Row
        connection = await connect(remainder)
        connection.row_factory = Row
        return vendor, connection



from supersql.core import Supersql
from supersql.column import Column
from supersql.table import Table
from supersql.query import Query
from supersql.results import Rows, Row


__version__ = "2020.1.24"
__all__ = ['Supersql', 'Column', 'Table', 'Query', 'Rows', 'Row',]



"""
Monty Python: Restaurant Sketch

Wife: It's nice here, isn't it?
Man: Oh, very good restaurant, three stars you know.
Wife: Really?
Man: Mmm...

Waiter: Good evening, sir! Good evening, madam! And may I say what a
    pleasure it is to see you here again, sir!

Man: Oh thank you. Well there you are dear. Have a look there, anything
    you like. The boeuf en croute is fantastic.

Waiter: Oh if I may suggest, sir ... the pheasant à la reine, the sauce
    is one of the chefs most famous creations.

Man: Em... that sounds good. Anyway just have a look... take your time. Oh, er
    by the way - got a bit of a dirty fork, could you ... er·.. get me another one?

Waiter: I beg your pardon.

Man: Oh it's nothing ... er, I've got a fork a little bit dirty. Could
    you get me another one? Thank you.

Waiter: Oh ... sir, 1 do apologize.
Man: Oh, no need to apologize, it doesn't worry me.
Waiter: Oh no, no, no, I do apologize. I will fetch the head waiter immediatement.
Man: Oh, there's no need to do that!
Waiter: Oh, no no... I'm sure the head waiter, he will want to apologize to you
    himself. I will fetch him at once.

Wife: Well, you certainly get good service here.
Man: They really look after you... yes.
"""


SYM = '--?--'


class Column():
    def __init__(self, col: str, table = None):
        self._name = col
        self._table = table
        self._sql = None
        self._alias = None
        self._arg = None

    def __eq__(self, v: any) -> 'Column':
        self._sql = f'{self._name} = {SYM}'
        self._arg = v
        return self

    def __ge__(self, v: any) -> 'Column':
        self._sql = f'{self._name} >= {SYM}'
        self._arg = v
        return self

    def __gt__(self, v: any) -> 'Column':
        self._sql = f'{self._name} > {SYM}'
        self._arg = v
        return self

    def __le__(self, v: any) -> 'Column':
        self._sql = f'{self._name} <= {SYM}'
        self._arg = v
        return self

    def __lt__(self, v: any) -> 'Column':
        self._sql = f'{self._name} < {SYM}'
        self._arg = v
        return self

    def __mod__(self, v: any) -> 'Column':
        self._sql = f'{self._name} LIKE {SYM}'
        self._arg = f'%{v}%'
        return self

    def __ne__(self, v: any) -> 'Column':
        self._sql = f'{self._name} <> {SYM}'
        self._arg = v
        return self

    def __neg__(self) -> 'Column':
        self._sql = f'{self._name} DESC'
        return self

    def __pos__(self) -> 'Column':
        self._sql = f'{self._name} ASC'
        return self

    def __str__(self) -> str:
        table_alias = self._table.__alias__ if self._table else None
        prefix = f'{table_alias}.' if table_alias else ''
        if self._alias:
            return f'{prefix}{self._name} AS {self._alias}'
        return f'{prefix}{self._name}'

    def AS(self, alias: str):
        self._alias = alias
        return self

    def LIKE(self, v, mask = '%%'):
        """
        This is to control the position of the '%' operator, something that
        column % val is too limited to do
        """
        if mask not in ['%%', '%-', '-%', '%_', '_%']:
            raise ValueError('invalid LIKE mask provided')
        self._sql = f'{self._name} LIKE {SYM}'
        val = f'{v}'
        if mask.startswith('%'):
            val = f'%{val}'
        if mask.endswith('%'):
            val = f'{val}%'
        self._arg = val
        return self

    @staticmethod
    def COERCE(f: 'Column'):
        if not isinstance(f, (str, Column, int, float, bool)):
            raise ValueError(f'''SELECT command accepts only <type, 'Column'> and Basic Data Types (int, float, bool..)''')
        return str(f)

    def QUOTE(param, stringify=False):
        if isinstance(param, str):
            return f"'{param}'"
        return param



from typing import List, Union
from asyncpg import Record


class Row():
    def __init__(self, current):
        self._current = current

    def __getattr__(self, k: str):
        return self._data.get(k)

    def column(self, k: Union[str, int]):
        return self._current[k]


class Rows():
    def __init__(self, rows: list):
        self._data = rows
        self._rows = [Row(row) for row in rows]
    
    def __bool__(self):
        return bool(self._data)

    def row(self, n: int):
        n = n - 1
        if n < 0: raise IndexError('Only positive row counts please...')
        return self._rows[n - 1]

    @property
    def rows(self) -> List[Row]:
        return self._data


# row = rows.row(5) -> Row
# rows = rows.rows[:3] -> List[Row]
# row.first_name -> str



"""
Head Waiter: Excuse me monsieur and madame. (examines the fork) It's filthy,
    Gaston ... find out who washed this up, and give them their cards immediately.

Man: Oh, no, no.

Head Waiter: Better still, we can't afford to take any chances, sack
    the entire washing-up staff.

Man: No, look I don't want to make any trouble.

Head Waiter: Oh, no please, no trouble. It's quite right that you
    should point these kind of things out. Gaston, tell the manager what has happened immediately! (The Waiter runs off)

Man: Oh, no I don't want to cause any fuss.

Head Waiter: Please, it's no fuss. I quite simply wish to ensure
    that nothing interferes with your complete enjoyment of the meal.

Man: Oh I'm sure it won't, it was only a dirty fork.

Head Waiter: I know. And I'm sorry, bitterly sorry, but I know that... no
    apologies I can make can alter the fact that in our restaurant you have
    been given a dirty, filthy, smelly piece of cutlery...

Man: It wasn't smelly.

Head Waiter: It was smelly, and obscene and disgusting and I hate it, I hate
    it ,.. nasty, grubby, dirty, mingy, scrubby little
    fork. Oh ... oh . . . oh . . . 
    (runs off in a passion as the manager comes to the table)
"""


from .column import Column


class Table(object):
    def __init__(self, table: str):
        self.__tn__ = table
        self.__alias__ = None

    def __getattr__(self, name: str) -> Column:
        return Column(name, self)
    
    def __str__(self) -> str:
        if self.__alias__:
            return f'{self.__tn__} AS {self.__alias__}'
        return self.__tn__

    def AS(self, alias: str):
        self.__alias__ = alias
        return self

    @staticmethod
    def COERCE(f: 'Table'):
        if not isinstance(f, (str, Table)):
            raise ValueError(f'''SELECT command accepts only {type('str')} and <type, 'Table'>''')
        return str(f)



"""
"""
from typing import List, Tuple, TYPE_CHECKING

from .constants import *
from .column import Column
from .results import Rows
from .table import Table


if TYPE_CHECKING:
    from supersql import Supersql  # pragma: no cover


class Query():
    def __init__(self, engine: 'Supersql'):
        self._engine = engine
        self._sql = []
        self._args = []
        self._vals = []
        self._zero = None

    def __str__(self) -> str:
        sql = ' '.join(s(False) for s in self._sql)
        return sql

    def _conditional(self, condition: Column, param: any, command: str):
        print(str(condition) * 40)
        if(isinstance(condition, str)):
            if not param:
                raise ValueError('''
                    To prevent SQL Injection please use parameterized query with string
                    or use Supersql <type, 'Column'> syntax.
                ''')
        elif not isinstance(condition, Column):
            raise ValueError(f'''
                Supersql {command} command only accepts <type, 'Column'> or a parameterized
                string and param second argument.
            ''')
        def _(xpand = False):
            if isinstance(condition, str):
                self._args.append(Column.QUOTE(param))
                sql = f'{command} {condition}'
            else:
                self._args.append(condition._arg)
                vendor = self._engine._vendor
                if vendor == POSTGRES: placeholder = f'${len(self._args)}'
                else: placeholder = '?'
                column_sql = condition._sql.replace('--?--', placeholder)
                sql = f'''{command} {column_sql}'''
            return sql
        self._sql.append(_)
        return self

    async def go(self):
        sql = ' '.join(s(True) for s in self._sql)
        zero = self._zero
        pooled = self._engine._pooled
        ispg = self._engine.vendor == POSTGRES
        connexion = self._engine._connexion

        async def go(conn):
            if zero == SELECT:
                if ispg:
                    results = await conn.fetch(sql, *self._args)
                else:
                    results = await conn.execute_fetchall(sql, self._args)
                return Rows(results)

            if len(self._vals) > 1:
                results = await conn.executemany(sql, tuple(self._vals))
                print('these are the results: ', results)
            else:
                results = await conn.execute(sql, *self._args)
            return Rows(results or [])

        if ispg and pooled:
            async with connexion.acquire() as connection:
                return await go(connection)
        else: return await go(connexion) # sqlite then until more databaseas added

    def _limit_offset(self, value: int, command: str):
        if not isinstance(value, int):
            raise ValueError(f'{command} only accepts integer values')
        self._sql.append(lambda xp: f'''{command} {value}''')
        return self

    def sql(self, unsafe = False) -> str:
        vendor = self._engine.vendor
        sql = str(self)
        if unsafe:
            if vendor != POSTGRES:
                for arg in self._args:
                    sql = sql.replace('?', arg, 1)
            else:
                for pos, arg in enumerate(self._args):
                    sql = sql.replace(f'${pos + 1}', arg, 1)
        return sql

    def AND(self, condition: Column | str, param = None):
        return self._conditional(condition, param, AND)

    def ASC(self):
        def _(xp):
            return f'ASC'
        self._sql.append(_)
        return self

    def DESC(self):
        def _(xp):
            return f'DESC'
        self._sql.append(_)
        return self

    def DELETE(self):
        self._sql.append(lambda xp: 'DELETE')
        return self

    def FROM(self, *tables: List[str | int]):
        def _(xp = ...):
            _tables = [Table.COERCE(t) for t in tables]
            return f'''FROM {', '.join(_tables)}'''
        self._sql.append(_)
        return self

    def INTO(self, table: Table):
        try: self._sql[-2]
        except: self._sql.append(lambda xp: f'INTO {table}')
        else: self._sql[-2] = lambda xp: f'INSERT INTO {table}'
        return self

    def INSERT(self, *columns: Column):
        self._zero = self._zero or INSERT
        def _(xp = ...):
            return f'''({', '.join(str(column) for column in columns)})'''
        self._sql.extend([lambda xp: '--', _])
        return self

    def INSERT_INTO(self, table: Table, columns: List[Column]):
        self._zero = self._zero or INSERT
        def _(xp = ...):
            return f'''INSERT INTO {table} ({', '.join(str(column) for column in columns)})'''
        self._sql.append(_)
        return self

    def LIMIT(self, limit: int):
        return self._limit_offset(limit, LIMIT)

    def OFFSET(self, offset: int):
        return self._limit_offset(offset, OFFSET)

    def OR(self, condition: Column | str, param = None):
        return self._conditional(condition, param, OR)

    def ORDER_BY(self, column: Column):
        def _(xp = ...):
            return f'''ORDER BY {column._sql or column}'''
        self._sql.append(_)
        return self

    def RETURNING(self, column: Column):
        def _(xp = ...):
            return f'RETURNING {column}'
        self._sql.append(_)
        return self

    def SELECT(self, *columns) -> 'Query':
        """Pythonic interface to SQL SELECT allowing python
        to be used to build SQL queries.

        Parameters:
            columns (List[str | int]): Variable args param that accepts either a string of
            Supersql Column Type.
            Raises an error if a type other than str | Column is used.
        """
        self._zero = self._zero or SELECT
        def _(xp = ...):
            _columns = [Column.COERCE(f) for f in columns]
            return f'''SELECT {', '.join(_columns)}''' if _columns else 'SELECT *'
        self._sql.append(_)
        return self

    def SET(self, *columns: Column):
        def _(xpand = False):
            vendor = self._engine._vendor
            statements = []
            for column in columns:
                self._args.append(column._arg)
                if vendor == POSTGRES: placeholder = f'${len(self._args)}'
                else: placeholder = '?'
                column_sql = column._sql.replace('--?--', placeholder)
                statements.append(column_sql)
            return f"SET {', '.join(statements)}"
        self._sql.append(_)
        return self

    def UPDATE(self, table: Table):
        self._zero = self._zero or UPDATE
        self._sql.append(lambda xp: f'''UPDATE {table}''')
        return self

    def VALUES(self, *matrix: Tuple[any]):
        msg = 'VALUES expects a tuple if inserting a row or multiple tuples if inserting multiple rows'
        homogeneous = set()
        for values in matrix:
            if not isinstance(values, (tuple, list)):
                raise ValueError(msg)
            homogeneous.add(len(values))
        if len(homogeneous) > 1:
            raise SyntaxError('VALUES matrix has tuples of different lengths')
        def _(xpand = False):
            vendor = self._engine._vendor
            count = homogeneous.copy().pop()
            if vendor == POSTGRES:
                placeholders = [f'${pos + 1}' for pos in range(count)]
            else:
                placeholders = ['?'] * count
            if xpand: self._vals.extend(matrix)
            return f'''VALUES ({', '.join(placeholders)})'''
        self._sql.append(_)
        return self

    def WHERE(self, condition: Column | str, param = None):
        return self._conditional(condition, param, WHERE)



AND = 'AND'
FROM = 'FROM'
INSERT = 'INSERT'
LIKE = 'LIKE'
LIMIT = 'LIMIT'
OFFSET = 'OFFSET'
OR = 'OR'
SELECT = 'SELECT'
SET = 'SET'
UPDATE = 'UPDATE'
WHERE = 'WHERE'


DATABASES = ['postgres', 'sqlite']

POSTGRES = 'postgres'
POSTGRES_LIVE_DSN = 'postgres://microstudian:eldorad0@localhost:5432/microstudio'

SQLITE = 'sqlite'
SQLITE_LIVE_DSN = 'sqlite://amebo.db'


from typing import List

from .helpers import Connexer
from .query import Query


class Supersql(object):
    def __init__(self, dsn: str):
        self._dsn = dsn
        self._vendor = None
        self._connexion = None
        self._pooled = False

    async def connect(self, pooled = False):
        self._pooled = pooled
        vendor, connector = await Connexer(self._dsn, self._pooled)
        self._connexion = connector
        self._vendor = vendor

    async def disconnect(self):
        assert self._connexion
        connexion = self._connexion
        await connexion.close()
        self._connexion = None

    def connects(self):
        """connect synchronous"""
        raise NotImplementedError

    def disconnects(self):
        """disconnect synchronous"""
        raise NotImplementedError

    @property
    def vendor(self) -> str:
        return self._vendor

    @vendor.setter
    def vendor(self, value: str):
        raise ValueError('To set Supersql vendor kindly provide a DSN at initialization')

    @property
    def query(self) -> Query:
        """Initializes Query and proxies Query.SELECT method directly"""
        return Query(self)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['supersql']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['supersql = supersql:main']}

setup_kwargs = {
    'name': 'supersql',
    'version': '2023.1.24',
    'description': 'Thin wrapper on top of SQL that enables you write SQL code in python easily',
    'long_description': 'Supersql Library\n================\nThere are many great database tools for python (i.e. databases, SQLAlchemy, PeeWee etc.) - **but there is no Python tool for databases.**\n\nIn addition you might have come to the same realisation and thinking the following:\n\n1. But we don\'t want to use an ORM\n\n2. Why can\'t we get a low level pythonic, powerful SQL api with with semantic interaction primitives\n\n3. Async and Sync support should be supported\n\nSupersql checks all those boxes and more. It is a python superset of SQL - allowing you leverage the full power of python to\nwrite advanced SQL queries.\n\n&nbsp;\n\n**Tutorial**: [Open Documentation](https://rayattack.github.io/supersql/)\n\n**Requirements**: Python 3.6+\n\n&nbsp;\n\n\n### NOTE: Still Very Much In Development\n\n```sql\nSELECT * FROM customers ORDER BY last_name ASC LIMIT 5\n```\n\n\n```py\n# query.SELECT(\'*\') is the same as query.SELECT() or query.SELECT(customers)\nquery.SELECT().FROM(customers).ORDER_BY(-customers.last_name).LIMIT(5)\n```\n\n&nbsp;\n\n## Why?\nLet\'s be honest:\n\n1. Writing sql templates using string formatting is really painful.\n2. Sometimes an ORM is not what you need, and whilst the new\n`f strings` in python solve a lot of problems, complex SQL templating is not of\nthem.\n\n3. Supersql makes it super simple to connect to and start querying a database in python.\n\n&nbsp;\n\nLet the code do the explanation:\n```py\n\nfrom supersql import Query\n\n\nquery = Query(\'postgres://user:password@hostname:5432/database\')\n\n\n# Without table schema discovery/reflection i.e. using strings -- NOT OPTIMAL\nresults = query.SELECT(\n        \'first_name\', \'last_name\', \'email\'\n    ).FROM(\n        \'employees\'\n    ).WHERE(\'email = someone@example.com\').execute()\n\nfor result in results:\n    print(result)\n\n\n# reflect table schema and fields into a python object for easy querying\nemps = query.database.table(\'employees\')\n\nrecords = query.SELECT(\n        emps.first_name, emps.last_name, emps.email\n    ).FROM(\n        emps\n    ).WHERE(emps.email == \'someone@example.com\').execute()\n```\n\n&nbsp;\n\nWhat about support for Code First flows? Also supported using Table objects\n```py\nfrom supersql import Table, Varchar, Date, Smallint\n\nclass Employee(Table):\n    """\n    SuperSQL is not an ORM. Table only helps you avoid magic\n    literals in your code. SuperSQL is not an ORM\n    """\n    __pk__ = (\'email\', \'identifier\')\n\n    identifier = Varchar()\n    email = Varchar(required=True, unique=None, length=25)\n    age = Smallint()\n    first_name = String(required=True)\n    last_name = String(25)\n    created_on = Date()\n\n\n# Now lets try again\nemp = Employee()\nresults = query.SELECT(\n    emp.first_name, emp.last_name, emp.email\n).FROM(emp).WHERE(\n    emp.email == \'someone@example.com\'\n).execute()\n```\n\n\n&nbsp;\n\n\n**Note**\n---\n**Supersql is not an ORM so there is no magic Table.save() Table.find() features nor will they ever be supported.**\nThe `Table` class is provided only to help with magic literal elimination from your codebase i.e. a query helper and nothing more.\n\n---\n',
    'author': 'Tersoo Ortserga',
    'author_email': 'codesage@live.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
